# -*- coding: utf-8 -*-

import json
from datetime import datetime
from logging import Logger

from cfgtools import Configuration
from typing import (
    Any,
    Dict,
    List,
    Optional
)
from xml.etree import cElementTree as eTree

import sysactions
from bustoken import TK_REMOTE_ORDER_CHECK_RUPTURA_DIFF_ITEMS
from msgbus import MBEasyContext, TK_SYS_ACK, FM_PARAM

from rupture_events import RuptureEvents, RuptureEventTypes
from rupture_repository import RuptureRepository
from model import RuptureItem, RuptureProduct

AUTOMATIC_RUPTURE_AFTER_SALE = "AUTOMATIC_RUPTURE_AFTER_SALE"


class RuptureService(object):
    def __init__(self, mb_context, logger, rupture_repository, configs):
        # type: (MBEasyContext, Logger, RuptureRepository, Configuration) -> None
        super(RuptureService, self).__init__()

        sysactions.mbcontext = mb_context
        self.mb_context = mb_context
        self.logger = logger
        self.rupture_repository = rupture_repository
        self._start = configs.find_value("Rupture.CleanRuptureTimeWindow.StartTime", None)
        self._end = configs.find_value("Rupture.CleanRuptureTimeWindow.EndTime", None)
        self._interval = configs.find_value("Rupture.CleanRuptureTimeWindow.Interval", "00:00:30")
        self._is_clean_enabled = configs.find_value("Rupture.CleanRuptureTimeWindow.Enabled", "false").lower() == "true"
        self._daily_dishes_event_types = ['STORE_ORDER', 'PAID']
        self._daily_dishes_products_cache = None
        self._last_auto_clean_event = None

    def load_daily_dishes_products(self):
        self._daily_dishes_products_cache = self.rupture_repository.get_all_daily_dishes()

    def get_enabled_items(self):
        self.logger.info("[START] Getting enabled items")
        enabled_items = self.rupture_repository.get_enabled_items()
        self.logger.info("[ END ] Getting enabled items")

        return {x.product_code: x.product_name for x in enabled_items}

    def get_enabled_items_list(self):
        self.logger.info("[START] Getting enabled items")
        enabled_items = self.rupture_repository.get_enabled_items()
        self.logger.info("[ END ] Getting enabled items")

        return self.get_rupture_products(enabled_items)

    def get_available_daily_dishes(
        self,
        exclude_daily_dishes    # type: List[str]
    ):
        # type: (...) -> List[RuptureProduct]

        self.logger.info("[START] Getting rupture by quantity list")
        rupture_by_qty_list = self.rupture_repository.get_available_daily_dishes_items(exclude_daily_dishes)
        self.logger.info("[END] Getting rupture by quantity list")

        return self.get_rupture_products(rupture_by_qty_list)

    def get_daily_dishes(
        self
    ):
        # type: () -> Dict[str, Any]

        self.logger.info("[START] Getting rupture daily dishes")

        all_disabled_items = []

        database_daily_dishes_ruptured = self.rupture_repository.get_daily_dishes_items()
        daily_dishes_ruptured = self.get_rupture_products(database_daily_dishes_ruptured)
        all_disabled_items.extend(daily_dishes_ruptured)

        items_ruptured = self.rupture_repository.get_disabled_items()
        all_disabled_items.extend(items_ruptured)

        disabled_product_codes = [str(dish.product_code) for dish in all_disabled_items]
        available_daily_dishes = self.get_available_daily_dishes(disabled_product_codes)

        daily_dishes = {
            "available": available_daily_dishes,
            "quantity_set": daily_dishes_ruptured
        }

        self.logger.info("[END] Getting rupture daily dishes")

        return daily_dishes

    def get_disabled_items(self):
        self.logger.info("[START] Getting disabled items")
        disabled_items = self.rupture_repository.get_disabled_items()
        self.logger.info("[ END ] Getting disabled items")

        return {x.product_code: x.product_name for x in disabled_items}

    def get_disabled_items_list(self):
        self.logger.info("[START] Getting disabled items")
        disabled_items = self.rupture_repository.get_disabled_items()
        self.logger.info("[ END ] Getting disabled items")

        return self.get_rupture_products(disabled_items)

    def get_rupture_products(self, items):
        # type: (List[RuptureItem]) -> List[RuptureProduct]
        item_list = []
        for item in items:
            product_code = item.product_code
            product_name = item.product_name
            products_composition = [item.product_part_code] if item.product_part_code else []
            product_quantity = item.product_quantity
            ruptured_product = RuptureProduct(
                product_code=product_code,
                product_name=product_name,
                products_composition=products_composition,
                product_quantity=product_quantity
            )

            if product_quantity == 0:
                delattr(ruptured_product, "product_quantity")

            if not next((x for x in item_list if x.product_code == product_code), None):
                item_list.append(ruptured_product)
            else:
                for list_item in item_list:
                    if self._must_add_product_composition(list_item, product_code, products_composition):
                        list_item.products_composition.extend(products_composition)
                        break
        return item_list

    @staticmethod
    def _must_add_product_composition(list_item, product_code, products_composition):
        if list_item.product_code != product_code:
            return False

        return products_composition is not None and products_composition not in list_item.products_composition

    def set_last_clean_event(
        self,
        period,  # type: str
    ):
        # type: (...) -> None
        self._last_auto_clean_event = period

    def check_items(self, changes_dict, session_id, all_items):
        message = "[START] Checking items in rupture. SessionId #{} - Items: {}"
        self.logger.info(message.format(session_id, changes_dict))

        items_to_disable = changes_dict.get("disabled") or []

        if not items_to_disable:
            self.logger.info("Has no items in rupture")
            return

        disabled_items = [] if all_items == "True" else list(self.get_disabled_items().iterkeys())
        filtered_items = set(map(lambda y: y["product_code"], items_to_disable)).difference(disabled_items)

        data = str(",".join(filtered_items) + "|" + ",".join(""))

        msg = self.mb_context.MB_EasySendMessage(
            dest_name="RemoteOrder",
            token=TK_REMOTE_ORDER_CHECK_RUPTURA_DIFF_ITEMS,
            format=FM_PARAM,
            data=data,
            timeout=180 * 1000000
        )
        if msg.token == TK_SYS_ACK:
            self.logger.info("[END] Checking items in rupture")
            return "" if msg.data == "" else msg.data.split(",")

        raise Exception("Cannot obtain remote order rupture diff items. Response: {}".format(msg.data))

    def update_items(self, changes_json, session_id, is_json=True):
        self.logger.info("[START] Updating items")

        changes_dict = json.loads(changes_json) if is_json else changes_json
        items_to_enable = changes_dict.get("enabled", [])
        items_to_disable = changes_dict.get("disabled", [])
        disabled_items = self.get_disabled_items()

        self.rupture_repository.insert_rupture_update(session_id, items_to_disable, items_to_enable, disabled_items)

        event_data = str(datetime.now().date())
        self.rupture_repository.insert_event_state(event_data)

        snapshot_id = self.create_full_snapshot()

        self._check_rupture_by_quantity(
            items_to_disable=items_to_disable,
            user_id=session_id
        )

        self.logger.info("[END] Updating items")

        return snapshot_id

    def update_daily_dishes(
        self,
        changes_json,   # type: str
        user_id         # type: str
    ):
        # type: (...) -> None

        self.logger.info("[START] Updating daily dishes")

        dishes_update = json.loads(changes_json)
        dishes_to_set_quantity = dishes_update.get("set", [])
        dishes_to_unset_quantity = dishes_update.get("unset", [])
        self.rupture_repository.insert_daily_dishes_update(
            dishes_to_set_quantity=dishes_to_set_quantity,
            dishes_to_unset_quantity=dishes_to_unset_quantity,
            user_id=user_id
        )

        self._check_rupture(
            items_to_enable=dishes_to_set_quantity,
            user_id=user_id
        )

        update_dishes = dict()
        remove_dishes = []
        for dish in dishes_to_set_quantity:
            if dish["product_quantity"]:
                update_dishes[dish["product_code"]] = dish["product_quantity"]
            else:
                remove_dishes.append(dish["product_code"])

        remove_dishes.extend([d["product_code"] for d in dishes_to_unset_quantity])

        self._daily_dishes_products_cache.update(update_dishes)
        for product_code in remove_dishes:
            self._daily_dishes_products_cache.pop(product_code)

        self.logger.info("[END] Updating daily dishes")

    def get_clean_time(self):
        return {
            "start": self._start,
            "end": self._end,
            "interval": self._interval
        }

    def create_full_snapshot(self):
        self.logger.info("[START] Creating rupture full snapshot")

        disabled_items = list(self.get_disabled_items().iterkeys())
        self.rupture_repository.mark_pending_ruptures_as_processed()
        snapshot_id = self.rupture_repository.insert_rupture_snapshot(disabled_items)
        self.rupture_repository.insert_rupture_current_state(snapshot_id)

        self._notify_rupture_changed(disabled_items, snapshot_id)

        self.logger.info("[ END ] Creating rupture full snapshot")

        return snapshot_id

    def clean_rupture_items(self, evt_type=None):
        if evt_type and evt_type in [RuptureEventTypes.CleanRuptureThread, RuptureEventTypes.CleanRuptureOnOpenDay]:
            event = 'AUTO_CLEAN_OPEN_DAY' if evt_type == RuptureEventTypes.CleanRuptureOnOpenDay else 'AUTO_CLEAN_EVENT'
            create_new_snapshot = self._clean_rupture_by_event(event_type=event)
            if create_new_snapshot:
                self.create_full_snapshot()
            self.rupture_repository.mark_clean_rupture_event_as_processed()
        elif evt_type and evt_type != RuptureEvents.CleanRupture:
            self._clean_rupture_by_button(evt_type)
            self.create_full_snapshot()
            self.rupture_repository.mark_clean_rupture_event_as_processed()

    def clean_rupture_with_catalog_update(self):
        create_new_snapshot = self._clean_rupture_by_event('AUTO_CLEAN_CATALOG_UPDATE')
        if create_new_snapshot:
            self.create_full_snapshot()
        self.rupture_repository.mark_clean_rupture_event_as_processed()

    def _notify_rupture_changed(self, disabled_items, snapshot_id):
        params = {
            "disabled_items": disabled_items,
            "snapshot_id": snapshot_id
        }
        self.mb_context.MB_EasyEvtSend("RupturaDataUpdated", "", json.dumps(params))
        self.logger.info("[RuptureDataUpdated] notified!")

    def clean_rupture_totem(self):
        try:
            pos_list = sysactions.get_poslist()
            for pos in pos_list:
                sysactions.set_custom(pos, "RUPTURA_ITEMS", "")
        except BaseException as e:
            self.logger.info('[ruptureItems] Exception {}'.format(e))

    def _clean_rupture_by_event(self, event_type='AUTO_CLEAN_EVENT'):
        # type: (Optional[str]) -> bool

        self.logger.info("[START] Cleaning all rupture products by event")

        if event_type == 'AUTO_CLEAN_OPEN_DAY':
            should_run = self._is_time_available()
            not_open_day_event = should_run and not self.rupture_repository.check_if_has_open_day_event()
            need_to_clean_ruptures = not_open_day_event and self.rupture_repository.check_if_need_to_clean_ruptures()
        else:
            need_to_clean_ruptures = self.rupture_repository.check_if_need_to_clean_ruptures()

        if not need_to_clean_ruptures:
            self.logger.info("[END] Cleaning all rupture products by event. No Ruptures to clean!")
            return False

        self.rupture_repository.clean_rupture_items(event_type)
        self.clean_rupture_totem()
        self.rupture_repository.mark_clean_rupture_event_as_processed()

        self.logger.info("[END] Cleaning all rupture products by event. Ruptures cleaned!")
        return True

    def _is_time_available(
        self,
    ):
        # type:  (...) -> bool
        now = datetime.now().time().isoformat()
        today = datetime.today().strftime("%Y%m%d")
        available_business_time = True
        if self._last_auto_clean_event:
            available_business_time = today != self._last_auto_clean_event
        return self._start and self._start <= now and available_business_time and self._is_clean_enabled

    def _clean_rupture_by_button(self, evt_type):
        self.logger.info("[START] Cleaning all rupture products by button")
        model = sysactions.get_model(evt_type)

        manager_id = sysactions.get_custom(model, "Last Manager ID")
        enable_session_id = "{} / {} / {}".format(manager_id, sysactions.get_operator_session(model), "CLEAN_BUTTON")

        self.rupture_repository.clean_rupture_items_by_btn(enable_session_id)
        self.logger.info("[ END ] Cleaning all rupture products by button")

    def _send_to_rupture(
        self,
        rupture_items,  # type: str
    ):
        # type: (...) -> None

        daily_dishes = self.get_daily_dishes()
        all_daily_dishes = daily_dishes["available"] + daily_dishes["quantity_set"]
        product_target = []
        for p in all_daily_dishes:
            if p.product_code in rupture_items:
                product_target.append(
                    {
                        'product_code': p.product_code,
                        'product_name': p.product_name,
                        'products_composition': p.products_composition
                    }
                )

        if len(product_target) > 0:
            self.update_items(
                changes_json={'disabled': product_target, 'enabled': []},
                session_id=AUTOMATIC_RUPTURE_AFTER_SALE,
                is_json=False
            )

    def _update_daily_dishes_quantity(
        self,
        product_code,   # type: str
        quantity,       # type: int
    ):
        # type: (...) -> List[str]

        quantity = max(quantity, 0)
        user = None if quantity > 0 else AUTOMATIC_RUPTURE_AFTER_SALE
        self.rupture_repository.change_dishes_quantity(
            product_code=product_code,
            quantity=quantity,
            user=user
        )

        if quantity > 0:
            self._daily_dishes_products_cache[product_code] = quantity
            return []

        self._daily_dishes_products_cache.pop(product_code)
        return [product_code]

    @staticmethod
    def _must_verify_rupture_by_quantity(
        order  # type: eTree
    ):
        # type: (...) -> bool

        if order.find('Order').get('subType') in ('SPLIT_SERVICE', 'SLICED_SERVICE'):
            return False

        order_stored_times = map(
            lambda x: x.get('state'),
            order.findall('Order/StateHistory/State')
        ).count('STORED')

        is_table_order = order.find("Order/CustomOrderProperties/OrderProperty[@key='TABLE_ID']") is not None
        is_delivery_and_stored = order.find('Order').get('saleTypeDescr') == "DELIVERY" and order_stored_times > 0

        order_state = order.find("Order").get("state")
        if is_table_order or is_delivery_and_stored:
            return order_state == "STORED" and order_stored_times == 1

        return order_state == "PAID"

    def decrement_item_in_rupture_by_quantity(
        self,
        data,       # type: str
        event_type  # type: unicode
    ):
        # type: (...) -> None

        if event_type in self._daily_dishes_event_types:
            order_picture = eTree.XML(data)
            if not self._must_verify_rupture_by_quantity(order=order_picture):
                return

            rupture_items = []
            for sale_line in order_picture.findall(".//SaleLine"):
                part_code = sale_line.get("partCode")
                if part_code not in self._daily_dishes_products_cache:
                    continue

                sale_line_quantity = int(float(sale_line.get("multipliedQty")))
                if sale_line_quantity == 0:
                    continue

                new_quantity = self._daily_dishes_products_cache[part_code] - sale_line_quantity
                rupture_items.extend(
                    self._update_daily_dishes_quantity(
                        product_code=part_code,
                        quantity=new_quantity
                    )
                )

            if rupture_items:
                self._send_to_rupture(rupture_items=rupture_items)

    def _check_rupture_by_quantity(
        self,
        items_to_disable,   # type: List[Dict]
        user_id             # type: str
    ):
        # type: (...) -> None

        disabled_items = []
        for products in items_to_disable:
            disabled_items.extend(products["products_composition"])
            disabled_items.append(products["product_code"])

        rupture_by_qty_list = self.rupture_repository.get_daily_dishes_items()
        daily_dishes_id = [products.product_code for products in rupture_by_qty_list]

        remove_ids = list(set(disabled_items).intersection(daily_dishes_id))
        dishes_to_unset_quantity = [{"product_code": value} for value in remove_ids]

        if remove_ids:
            self.rupture_repository.insert_daily_dishes_update(
                dishes_to_set_quantity=[],
                dishes_to_unset_quantity=dishes_to_unset_quantity,
                user_id=user_id
            )

    def _check_rupture(
        self,
        items_to_enable,    # type: List[Dict]
        user_id             # type: str
    ):
        # type: (...) -> None

        enabled_items = []
        disabled_items = []
        for products in items_to_enable:
            if products["product_quantity"] == 0:
                disabled_items.append(products)
            else:
                enabled_items.append(products)

        changes_dict = {
            "disabled": disabled_items,
            "enabled": enabled_items
        }

        if enabled_items or disabled_items:
            self.update_items(
                changes_json=changes_dict,
                session_id=user_id,
                is_json=False
            )



# -*- coding: utf-8 -*-

import json
from logging import Logger

import bustoken
from messagehandler import EventHandler
from msgbus import MBEasyContext, TK_SYS_ACK, TK_SYS_NAK, FM_STRING

from rupture_events import RuptureEvents
from rupture_service import RuptureService


class RuptureEventHandler(EventHandler):
    def __init__(self, mb_context, logger, rupture_service):
        # type: (MBEasyContext, Logger, RuptureService) -> None
        super(RuptureEventHandler, self).__init__(mb_context)

        self.logger = logger
        self.rupture_service = rupture_service

    def get_handled_tokens(self):
        return [bustoken.TK_RUPTURA_GET_ENABLED,
                bustoken.TK_RUPTURA_GET_DISABLED,
                bustoken.TK_RUPTURA_UPDATE_ITEMS,
                bustoken.TK_RUPTURA_CHECK_ITEMS,
                bustoken.TK_RUPTURA_GET_CLEAN_TIME_WINDOW,
                bustoken.TK_RUPTURA_CLEAN_WITH_CATALOG_UPDATE,
                bustoken.TK_RUPTURA_GET_DAILY_DISHES,
                bustoken.TK_RUPTURA_UPDATE_DAILY_DISHES]

    def handle_message(self, msg):
        try:
            processing_message = "Start processing message: {}"

            if msg.token == bustoken.TK_RUPTURA_CHECK_ITEMS:
                self.logger.info(processing_message.format("TK_RUPTURA_CHECK_ITEMS"))

                params = msg.data.split('\0')
                changes_json, session_id = params[:2]

                all_items = "False"
                if len(params) == 3:
                    all_items = params[2]

                changes_dict = json.loads(changes_json or "{}")
                items_to_disable = self.rupture_service.check_items(changes_dict, session_id, all_items)
                data = json.dumps(items_to_disable) if items_to_disable else ""

                self._success_reply_message(msg, data)

            elif msg.token == bustoken.TK_RUPTURA_UPDATE_ITEMS:
                self._update_rupture_items(msg, processing_message)

            elif msg.token == bustoken.TK_RUPTURA_GET_ENABLED:
                self.logger.info(processing_message.format("TK_RUPTURA_GET_ENABLED"))

                if msg.data == 'UI':
                    items = self.rupture_service.get_enabled_items_list()
                    data = json.dumps(items, default=lambda o: o.__dict__, sort_keys=True)
                else:
                    data = json.dumps(self.rupture_service.get_enabled_items())

                self._success_reply_message(msg, data)

            elif msg.token == bustoken.TK_RUPTURA_GET_DAILY_DISHES:
                self.logger.info(processing_message.format("TK_RUPTURA_GET_DAILY_DISHES"))

                daily_dishes = self.rupture_service.get_daily_dishes()
                data = json.dumps(daily_dishes, default=lambda o: o.__dict__, sort_keys=True)

                self._success_reply_message(msg, data)

            elif msg.token == bustoken.TK_RUPTURA_GET_DISABLED:
                self.logger.info(processing_message.format("TK_RUPTURA_GET_DISABLED"))

                if msg.data == 'UI':
                    items = self.rupture_service.get_disabled_items_list()
                    data = json.dumps(items, default=lambda o: o.__dict__, sort_keys=True)
                else:
                    data = json.dumps(self.rupture_service.get_disabled_items())

                self._success_reply_message(msg, data)

            elif msg.token == bustoken.TK_RUPTURA_GET_CLEAN_TIME_WINDOW:
                self.logger.info(processing_message.format("TK_RUPTURA_GET_CLEAN_TIME_WINDOW"))
                data = json.dumps(self.rupture_service.get_clean_time())
                self._success_reply_message(msg, data)

            elif msg.token == bustoken.TK_RUPTURA_CLEAN_WITH_CATALOG_UPDATE:
                self.logger.info(processing_message.format("TK_RUPTURA_CLEAN_WITH_CATALOG_UPDATE"))
                self.rupture_service.clean_rupture_with_catalog_update()
                self._success_reply_message(msg)

            elif msg.token == bustoken.TK_RUPTURA_UPDATE_DAILY_DISHES:
                self._update_daily_dishes(
                    msg=msg,
                    processing_message=processing_message
                )

        except BaseException as ex:
            self.logger.exception("Error handling message")
            msg.token = TK_SYS_NAK
            self.mbcontext.MB_ReplyMessage(msg, FM_STRING, repr(ex))

    def _update_rupture_items(self, msg, processing_message):
        self.logger.info(processing_message.format("TK_RUPTURA_UPDATE_ITEMS"))

        changes_json, session_id, pos_id = msg.data.split('\0')
        self.rupture_service.update_items(changes_json, session_id)

        self._success_reply_message(msg)

    def _update_daily_dishes(
        self,
        msg,                # type: MBEasyContext
        processing_message  # type: str
    ):
        # type: (...) -> None

        self.logger.info(processing_message.format("TK_RUPTURA_UPDATE_DAILY_DISHES"))

        changes_json, user_id = msg.data.split('\0')
        self.rupture_service.update_daily_dishes(
            changes_json=changes_json,
            user_id=user_id
        )

        self._success_reply_message(msg)

    def handle_event(self, subject, evt_type, data, msg):
        try:
            self.logger.info("Start processing event: {}".format(subject))

            if subject == RuptureEvents.FullRuptureRequest:
                self.rupture_service.create_full_snapshot()
            elif subject == RuptureEvents.CleanRupture:
                self.rupture_service.clean_rupture_items(evt_type)
            elif subject == RuptureEvents.OrderModified:
                self.rupture_service.decrement_item_in_rupture_by_quantity(
                    data=data,
                    event_type=evt_type
                )

        except Exception as _:
            self.logger.exception("Error processing event: {}".format(subject))
        finally:
            self.logger.info("Finishing event: {}".format(subject))

    def terminate_event(self):
        # No need to stop anything into this component
        pass

    def _success_reply_message(self, msg, data=None):
        msg.token = TK_SYS_ACK
        if data:
            self.mbcontext.MB_ReplyMessage(msg, FM_STRING, data)
        else:
            self.mbcontext.MB_EasyReplyMessage(msg)



# inn.py - functions for handling Russian VAT numbers
# coding: utf-8
#
# Copyright (C) 2015 Arthur de Jong
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301 USA

"""ИНН (Идентификационный номер налогоплательщика, Russian tax identifier).

The Indentifikatzionny nomer nalogoplatel'shchika is a Russian tax
identification number that consists 10 digits for companies and 12 digits for
persons.

>>> validate('123456789047')
'123456789047'
>>> validate('1234567894')
'1234567894'
>>> validate('123456789037')
Traceback (most recent call last):
    ...
InvalidChecksum: ...
>>> validate('1234567895')
Traceback (most recent call last):
    ...
InvalidChecksum: ...
"""

from stdnum.exceptions import *
from stdnum.util import clean, isdigits


def compact(number):
    """Convert the number to the minimal representation. This strips the
    number of any valid separators and removes surrounding whitespace."""
    return clean(number, ' ').strip()


def calc_company_check_digit(number):
    """Calculate the check digit for the 10-digit ИНН for organisations."""
    weights = (2, 4, 10, 3, 5, 9, 4, 6, 8)
    return str(sum(w * int(n) for w, n in zip(weights, number)) % 11 % 10)


def calc_personal_check_digits(number):
    """Calculate the check digits for the 12-digit personal ИНН."""
    weights = (7, 2, 4, 10, 3, 5, 9, 4, 6, 8)
    d1 = str(sum(w * int(n) for w, n in zip(weights, number)) % 11 % 10)
    weights = (3, 7, 2, 4, 10, 3, 5, 9, 4, 6, 8)
    d2 = str(sum(w * int(n) for w, n in zip(weights, number[:10] + d1)) % 11 % 10)
    return d1 + d2


def validate(number):
    """Check if the number is a valid ИНН. This checks the length, formatting
    and check digit."""
    number = compact(number)
    if not isdigits(number):
        raise InvalidFormat()
    if len(number) == 10:
        if calc_company_check_digit(number) != number[-1]:
            raise InvalidChecksum()
    elif len(number) == 12:
        # persons
        if calc_personal_check_digits(number) != number[-2:]:
            raise InvalidChecksum()
    else:
        raise InvalidLength()
    return number


def is_valid(number):
    """Check if the number is a valid ИНН."""
    try:
        return bool(validate(number))
    except ValidationError:
        return False



# Copyright Jonathan Hartley 2013. BSD 3-Clause license, see LICENSE file.
import sys
from unittest import TestCase, main, skipUnless

try:
    from unittest.mock import patch, Mock
except ImportError:
    from mock import patch, Mock

from ..ansitowin32 import StreamWrapper
from ..initialise import init, just_fix_windows_console, _wipe_internal_state_for_tests
from .utils import osname, replace_by

orig_stdout = sys.stdout
orig_stderr = sys.stderr


class InitTest(TestCase):

    @skipUnless(sys.stdout.isatty(), "sys.stdout is not a tty")
    def setUp(self):
        # sanity check
        self.assertNotWrapped()

    def tearDown(self):
        _wipe_internal_state_for_tests()
        sys.stdout = orig_stdout
        sys.stderr = orig_stderr

    def assertWrapped(self):
        self.assertIsNot(sys.stdout, orig_stdout, 'stdout should be wrapped')
        self.assertIsNot(sys.stderr, orig_stderr, 'stderr should be wrapped')
        self.assertTrue(isinstance(sys.stdout, StreamWrapper),
            'bad stdout wrapper')
        self.assertTrue(isinstance(sys.stderr, StreamWrapper),
            'bad stderr wrapper')

    def assertNotWrapped(self):
        self.assertIs(sys.stdout, orig_stdout, 'stdout should not be wrapped')
        self.assertIs(sys.stderr, orig_stderr, 'stderr should not be wrapped')

    @patch('colorama.initialise.reset_all')
    @patch('colorama.ansitowin32.winapi_test', lambda *_: True)
    @patch('colorama.ansitowin32.enable_vt_processing', lambda *_: False)
    def testInitWrapsOnWindows(self, _):
        with osname("nt"):
            init()
            self.assertWrapped()

    @patch('colorama.initialise.reset_all')
    @patch('colorama.ansitowin32.winapi_test', lambda *_: False)
    def testInitDoesntWrapOnEmulatedWindows(self, _):
        with osname("nt"):
            init()
            self.assertNotWrapped()

    def testInitDoesntWrapOnNonWindows(self):
        with osname("posix"):
            init()
            self.assertNotWrapped()

    def testInitDoesntWrapIfNone(self):
        with replace_by(None):
            init()
            # We can't use assertNotWrapped here because replace_by(None)
            # changes stdout/stderr already.
            self.assertIsNone(sys.stdout)
            self.assertIsNone(sys.stderr)

    def testInitAutoresetOnWrapsOnAllPlatforms(self):
        with osname("posix"):
            init(autoreset=True)
            self.assertWrapped()

    def testInitWrapOffDoesntWrapOnWindows(self):
        with osname("nt"):
            init(wrap=False)
            self.assertNotWrapped()

    def testInitWrapOffIncompatibleWithAutoresetOn(self):
        self.assertRaises(ValueError, lambda: init(autoreset=True, wrap=False))

    @patch('colorama.win32.SetConsoleTextAttribute')
    @patch('colorama.initialise.AnsiToWin32')
    def testAutoResetPassedOn(self, mockATW32, _):
        with osname("nt"):
            init(autoreset=True)
            self.assertEqual(len(mockATW32.call_args_list), 2)
            self.assertEqual(mockATW32.call_args_list[1][1]['autoreset'], True)
            self.assertEqual(mockATW32.call_args_list[0][1]['autoreset'], True)

    @patch('colorama.initialise.AnsiToWin32')
    def testAutoResetChangeable(self, mockATW32):
        with osname("nt"):
            init()

            init(autoreset=True)
            self.assertEqual(len(mockATW32.call_args_list), 4)
            self.assertEqual(mockATW32.call_args_list[2][1]['autoreset'], True)
            self.assertEqual(mockATW32.call_args_list[3][1]['autoreset'], True)

            init()
            self.assertEqual(len(mockATW32.call_args_list), 6)
            self.assertEqual(
                mockATW32.call_args_list[4][1]['autoreset'], False)
            self.assertEqual(
                mockATW32.call_args_list[5][1]['autoreset'], False)


    @patch('colorama.initialise.atexit.register')
    def testAtexitRegisteredOnlyOnce(self, mockRegister):
        init()
        self.assertTrue(mockRegister.called)
        mockRegister.reset_mock()
        init()
        self.assertFalse(mockRegister.called)


class JustFixWindowsConsoleTest(TestCase):
    def _reset(self):
        _wipe_internal_state_for_tests()
        sys.stdout = orig_stdout
        sys.stderr = orig_stderr

    def tearDown(self):
        self._reset()

    @patch("colorama.ansitowin32.winapi_test", lambda: True)
    def testJustFixWindowsConsole(self):
        if sys.platform != "win32":
            # just_fix_windows_console should be a no-op
            just_fix_windows_console()
            self.assertIs(sys.stdout, orig_stdout)
            self.assertIs(sys.stderr, orig_stderr)
        else:
            def fake_std():
                # Emulate stdout=not a tty, stderr=tty
                # to check that we handle both cases correctly
                stdout = Mock()
                stdout.closed = False
                stdout.isatty.return_value = False
                stdout.fileno.return_value = 1
                sys.stdout = stdout

                stderr = Mock()
                stderr.closed = False
                stderr.isatty.return_value = True
                stderr.fileno.return_value = 2
                sys.stderr = stderr

            for native_ansi in [False, True]:
                with patch(
                    'colorama.ansitowin32.enable_vt_processing',
                    lambda *_: native_ansi
                ):
                    self._reset()
                    fake_std()

                    # Regular single-call test
                    prev_stdout = sys.stdout
                    prev_stderr = sys.stderr
                    just_fix_windows_console()
                    self.assertIs(sys.stdout, prev_stdout)
                    if native_ansi:
                        self.assertIs(sys.stderr, prev_stderr)
                    else:
                        self.assertIsNot(sys.stderr, prev_stderr)

                    # second call without resetting is always a no-op
                    prev_stdout = sys.stdout
                    prev_stderr = sys.stderr
                    just_fix_windows_console()
                    self.assertIs(sys.stdout, prev_stdout)
                    self.assertIs(sys.stderr, prev_stderr)

                    self._reset()
                    fake_std()

                    # If init() runs first, just_fix_windows_console should be a no-op
                    init()
                    prev_stdout = sys.stdout
                    prev_stderr = sys.stderr
                    just_fix_windows_console()
                    self.assertIs(prev_stdout, sys.stdout)
                    self.assertIs(prev_stderr, sys.stderr)


if __name__ == '__main__':
    main()



from typing import (
    Dict,
    List,
    Tuple,
)

from ._DeviceConfig import DeviceConfig


class Configs(object):

    def __init__(
        self,
        component_name,     # type: str
        required_services,  # type: str
        persistence_name,   # type: str
        sector_config,      # type: Dict[str, Dict[str, Tuple[str, Dict[int, int]]]]
        device_configs,     # type: List[DeviceConfig]
        brand_config,       # type: Dict[str, Dict[str, str]]
    ):
        # type: (...) -> None

        self.component_name = component_name
        self.required_services = required_services
        self.persistence_name = persistence_name
        self.sector_config = sector_config
        self.device_configs = device_configs
        self.brand_config = brand_config



from production.box import FilterDefaultQuantityIngredientsBox
from production.treebuilder import BoxConfiguration

from ._BoxBuilder import BoxBuilder


class FilterDefaultQuantityIngredientsBoxBuilder(BoxBuilder):
    def __init__(self, loggers):
        super(FilterDefaultQuantityIngredientsBoxBuilder, self).__init__(loggers)

    def build(self, config):
        # type: (BoxConfiguration) -> FilterDefaultQuantityIngredientsBox
        return FilterDefaultQuantityIngredientsBox(config.name, config.sons, self.get_logger(config))



import json

from sysactions import (
    action,
    show_any_dialog,
    show_messagebox,
    StopAction,
)

import get_products

from .. import logger


@action
def open_rupture(pos_id):
    try:
        enabled_ingredients = get_products.get_enabled_products()
        disabled_ingredients = get_products.get_disabled_products()

    except Exception as e:
        logger.exception('[rupture_items] Exception {}'.format(e))
        show_messagebox(
            posid=pos_id,
            message="$ERROR_GETTING_DATA",
            title="$INFORMATION",
            buttons="$EXIT"
        )

        raise StopAction()

    dialog_contents = {
        "type": "rupture",
        "availableItems": enabled_ingredients,
        "selectedItems": disabled_ingredients
    }

    info_text = json.dumps(dialog_contents)
    show_any_dialog(
        posid=pos_id,
        dlgtype="enableItemsDialog",
        message="",
        title="$MANAGE_STOCKOUT_ITEMS",
        infotext=info_text,
        buttons="",
        icon="",
        timeout=6000000,
        mask="",
        defvalue="",
        minvalue="",
        maxvalue="",
        contents="",
        asynch=False
    )
